from enum import Enum
import operator
import os
import pathlib
from typing import Callable, Sequence, Tuple, Union

from autoclick.types.library import (
    ReadablePath,
    ReadableFile,
    ReadableDir,
    WritablePath,
    WritableFile,
    WritableDir
)
from autoclick.validations import ValidationError, validation


class Comparison(Enum):
    LT = ("<", operator.lt)
    LE = ("<=", operator.le)
    GT = (">", operator.gt)
    GE = (">=", operator.ge)
    EQ = ("=", operator.eq)
    NE = ("!=", operator.ne)

    def __init__(self, symbol: str, fn: Callable):
        self.symbol = symbol
        self.fn = fn


class Defined:
    """
    Validate that some number of paramters are defined.

    Args:
        n: Number to compare against the number of defined parameters.
        cmp: Comparison.
    """
    def __init__(self, n: int = 1, cmp: Comparison = Comparison.GE):
        self.n = n
        self.cmp = cmp

    def __call__(self, **kwargs):
        defined = len(tuple(filter(None, kwargs.values())))
        if not self.cmp.fn(defined, self.n):
            raise ValidationError(
                f"Of the following parameters, the number defined must be " 
                f"{self.cmp.symbol} {self.n}: {', '.join(kwargs.keys())}"
            )


def defined_ge(n: int):
    return Defined(n, Comparison.GE)


class Mutex:
    def __init__(self, *groups: Union[int, Tuple[int, ...]], max_defined: int = 1):
        self.groups = groups
        self.max_defined = max_defined

    def __call__(self, **kwargs):
        args = list(kwargs.items())
        groups = self.groups or range(len(args))
        defined = []
        for group in groups:
            if isinstance(group, int):
                group = [group]
            for i in group:
                if args[i][1] is not None:
                    defined.append(group)
                    break
        if len(defined) > self.max_defined:
            group_str = ",".join(f"({','.join(str(i) for i in g)})" for g in defined)
            raise ValidationError(
                f"Values specified for > {self.max_defined} mutually exclusive groups: "
                f"{group_str}"
            )


class SequenceLength:
    """
    Validate that the length of a sequence is within certain bounds.

    Args:
        *lengths: either a single value, which specifies the exact length a sequence
            must have, or a min and max value.
    """
    def __init__(self, *lengths):
        if len(lengths) == 1:
            self.minlen = self.maxlen = lengths[0]
        else:
            self.minlen, self.maxlen = lengths
        if self.minlen > self.maxlen:
            raise ValueError("'minlen' must be <= 'maxlen'")

    def __call__(self, param_name: str, value: Sequence):
        if len(value) < self.minlen:
            raise ValidationError(
                f"Parameter {param_name} must have at least {self.minlen} elements."
            )
        if len(value) < self.maxlen:
            raise ValidationError(
                f"Parameter {param_name} can have at most {self.maxlen} elements."
            )


@validation(ReadablePath)
def readable_path(param_name: str, value: pathlib.Path):
    if not value.exists():
        raise ValidationError(
            f"Parameter {param_name} value {value} does not exist."
        )


@validation(ReadableFile, depends=(readable_path,))
def readable_file(param_name: str, value: pathlib.Path):
    if not value.is_file():
        raise ValidationError(
            f"Parameter {param_name} value {value} is not a file."
        )


@validation(ReadableDir, depends=(readable_path,))
def readable_dir(param_name: str, value: pathlib.Path):
    if not value.is_dir():
        raise ValidationError(
            f"Parameter {param_name} value {value} is not a directory."
        )


@validation(WritablePath)
def writable_path(param_name: str, value: pathlib.Path):
    existing = value
    while not existing.exists():
        if existing.parent:
            existing = existing.parent
        else:
            break
    if not os.access(existing, os.W_OK):
        raise ValidationError(
            f"Parameter {param_name} value {value} is not writable."
        )


@validation(WritableFile, depends=(writable_path,))
def writable_file(param_name: str, value: pathlib.Path):
    if value.exists() and not value.is_file():
        raise ValidationError(
            f"Parameter {param_name} value {value} exists and is not a file."
        )


@validation(WritableDir, depends=(writable_path,))
def writable_dir(param_name: str, value: pathlib.Path):
    if value.exists() and not value.is_dir():
        raise ValidationError(
            f"Parameter {param_name} value {value} exists and is not a directory."
        )



from pathlib import Path
import sys
from typing import Optional, Type

import click


def version_option(
    version: Optional[str] = None, module: Optional[str] = None, *param_decls,
    prog_name: Optional[str] = None, message: str = "{prog}, version {ver}",
    option_class: Type[click.Option] = click.Option, **kwargs
) -> click.Option:
    """Adds a ``--version`` option which immediately ends the program
    printing out the version number.  This is implemented as an eager
    option that prints the version and exits the program in the callback.

    Args:
        version: the version number to show. If not provided, attempts an auto
            discovery via setuptools.
        module:
        prog_name: the name of the program (defaults to autodetection)
        message: custom message to show instead of the default
            (``"{prog}, version {ver}"``)
        option_class:
        kwargs: everything else is forwarded to :func:`option`.
    """
    def callback(ctx, param, value):
        if not value or ctx.resilient_parsing:
            return
        ver = version or get_version(module or get_module())
        prog = prog_name or ctx.find_root().info_name
        click.echo(message.format(prog, ver), color=ctx.color)
        ctx.exit()

    kwargs.setdefault("is_flag", True)
    kwargs.setdefault("expose_value", False)
    kwargs.setdefault("is_eager", True)
    kwargs.setdefault("help", "Show the version and exit.")
    kwargs["callback"] = callback

    return option_class(*(param_decls or ("--version",)), **kwargs)


def get_module():
    if hasattr(sys, "_getframe"):
        return sys._getframe(1).f_globals.get("__name__")
    else:
        raise RuntimeError("Could not determine module")


def get_version(module):
    ver = None

    # Try pkg_resources
    try:
        import pkg_resources
    except ImportError:
        pass
    else:
        try:
            # Try get_distribution
            ver = pkg_resources.get_distribution(module).version
        except:
            # Fall back to looking for entry point
            for dist in pkg_resources.working_set:
                scripts = dist.get_entry_map().get("console_scripts") or {}
                for script_name, entry_point in scripts.items():
                    if entry_point.module_name == module:
                        ver = dist.version
                        break

    # Try pyproject.toml
    try:
        toml_mod = get_toml_parser()
    except ImportError:
        pass
    else:
        path = Path.cwd()
        while path:
            pyproj = path / "pyproject.toml"
            if pyproj.exists():
                with open(pyproj, "rt") as inp:
                    ver = toml_mod.parse(inp.read())["tool"]["poetry"]["version"]
            else:
                path = path.parent

    if ver is None:
        raise RuntimeError("Could not determine version")

    return ver


def get_toml_parser():
    try:
        import tomlkit
        return tomlkit
    except ImportError:
        pass

    import toml
    return toml



from abc import ABCMeta, abstractmethod
import functools
import inspect
from typing import (
    Any,
    Callable,
    Dict,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
    cast
)

import click

from autoclick.core import (
    DEC,
    BaseDecorator,
    ParameterInfo,
    SignatureError,
    apply_to_parsed_args
)
from autoclick.utils import get_dest_type


class TypeCollisionError(Exception):
    """Raised when a decorator is defined for a type for which a decorator of the
    same kind has already been defined.
    """


class Composite(BaseDecorator[DEC], metaclass=ABCMeta):
    """
    Represents a complex type that requires values from multiple parameters. A
    composite parameter is defined by annotating a class using the `composite_type`
    decorator, or by annotating a function with the `composite_factory` decorator.
    The parameters of the composite type's construtor (exluding `self`) or of the
    composite factory function are added to the command prior to argument parsing,
    and then they are replaced by an instance of the annotated class after parsing.

    Note that composite parameters cannot be nested, i.e. a parameter cannot be a
    list of composite types, and a composite type cannot itself have composite type
    parameters - either of these will raise a :class:`SignatureError`.

    Args:
        parameters_as_args: Whether to treat all parameters as Arguments regardless
            of whether they are optional or required.
        force_create: Always create an instance of the composite type, even if all
            the parameter values are `None`.
        ignore_type_args: Whether it is safe to ignore any type arguments when matching.
        kwargs: Keyword arguments passed to :class:`BaseDecorator` constructor.

    Kwargs:
        to_wrap: The function/class to wrap.
        keep_underscores: Whether underscores should be retained in option names
            (True) or converted to hyphens (False).
        short_names: Dictionary mapping parameter names to short names. If
            not specified, usage of short names depends on `infer_short_names`.
            Set a value to `None` to disable short name usage for a paramter.
        infer_short_names: Whether to infer short names from parameter names.
            See Details on the algorithm used to select the short name. If a
            parameter has a short name specified in `short_names` it overrides
            the inferred short name.
        option_order: Specify an order of option processing that is different
            from the order in the signature of the annotated function.
        types: Dict mapping parameter names to functions that perform type
            conversion. By default, the type of a parameter is inferred from
            its annotation.
        positionals_as_options: Whether to treat positional arguments as required
            options.
        conditionals: Dict mapping paramter names or tuples of parameter names to
            conditional functions or lists of conditinal functions.
        validations: Dict mapping paramter names or tuples of parameter names to
            validation functions or lists of validation functions.
        required: Sequence of required options. If not specified, only paramters
            without default values are required.
        show_defaults: Whether to show defaults in the help text.
        hidden: Sequence of hidden options. These options are not displayed in the
            help text.
        param_help: Dict mapping parameters to help strings. By default, these are
            extracted from the function docstring.
    """
    def __init__(
        self,
        force_create: bool = False,
        parameters_as_args: bool = False,
        ignore_type_args: bool = False,
        **kwargs
    ):
        self.force_create = force_create
        self._parameters_as_args = parameters_as_args
        self._ignore_type_args = ignore_type_args
        self._parameters = None
        super().__init__(**kwargs)

    @property
    @abstractmethod
    def _match_type(self) -> Callable:
        """
        """
        pass

    def _handle_parameter_info(self, param: ParameterInfo) -> bool:
        if param.extra_arguments or param.extra_kwargs:
            raise SignatureError(
                "CompositeType cannot have *args or **kwargs"
            )
        return super()._handle_parameter_info(param)

    def _create_decorator(self) -> DEC:
        self._parameters = self._get_parameter_info()
        if has_composite(self._match_type):
            raise TypeCollisionError(
                f"A composite for type {self._match_type} is already defined."
            )
        register_composite(self._match_type, self)
        if (
            self._ignore_type_args and
            hasattr(self._match_type, "__args__") and
            hasattr(self._match_type, "__origin__")
        ):
            register_composite(self._match_type.__origin__, self)

        return self._decorated

    def create_click_parameters(
        self,
        param: ParameterInfo,
        used_short_names: Set[str],
        add_prefixes: bool,
        hidden: bool,
        default_values: Dict[str, Any],
        help_text: str,
        option_class: Type[click.Option],
        argument_class: Type[click.Argument],
        force_positionals_as_options: bool = False
    ) -> Tuple[Sequence[click.Parameter], Callable[[dict], None]]:
        """
        Create the Click parameters for this composite's signature.

        Args:
            param:
            used_short_names:
            add_prefixes:
            hidden:
            default_values:
            help_text:
            option_class:
            argument_class:
            force_positionals_as_options:

        Returns:
             A tuple (click_parameters, callback), where click_parameters is a
             list of :class:`click.Option` or :class:`click.Argument` instances,
             and the callback is the function to be called with the actual parameter
             values after the command line is parsed. If `self.parameters_as_args` is
             True, a single :class:`click.Tuple` instance.
        """
        if self._parameters_as_args:
            param_decls = ["--{}".format(self._get_long_name(param.name))]

            short_name = self._get_short_name(param.name, used_short_names)
            if short_name:
                used_short_names.add(short_name)
                param_decls.append(f"-{short_name}")

            types = []
            default = []
            for param_name in self._option_order:
                composite_param = self._parameters[param_name]
                types.append(composite_param.click_type)
                default.append(default_values.get(param_name, composite_param.default))

            click_parameters = [
                option_class(
                    param_decls,
                    type=click.Tuple(types),
                    required=not param.optional,
                    default=default,
                    show_default=self._show_defaults,
                    nargs=len(types),
                    hidden=hidden,
                    is_flag=False,
                    multiple=False,
                    help=help_text
                )
            ]
        else:
            prefix = param.name if add_prefixes else None
            click_parameters = [
                self._create_click_parameter(
                    param=self._parameters[opt],
                    used_short_names=used_short_names,
                    option_class=option_class,
                    argument_class=argument_class,
                    long_name_prefix=prefix,
                    hidden=hidden,
                    default_values=default_values,
                    force_positionals_as_options=force_positionals_as_options
                )
                for opt in self._option_order
            ]

        callback = cast(
            Callable[[dict], None],
            functools.partial(self.handle_args, param=param, add_prefixes=add_prefixes)
        )

        return click_parameters, callback

    def handle_args(self, ctx: click.Context, param: ParameterInfo, add_prefixes: bool):
        if self._parameters_as_args:
            kwargs = dict(zip(self._option_order, ctx.params.pop(param.name, ())))
            apply_to_parsed_args(self._conditionals, kwargs, update=True)
            apply_to_parsed_args(self._validations, kwargs, update=False)
        else:
            kwargs = {}
            for composite_param_name in self._parameters.keys():
                if add_prefixes:
                    arg_name = f"{param.name}_{composite_param_name}"
                else:
                    arg_name = composite_param_name
                kwargs[composite_param_name] = ctx.params.pop(arg_name, None)

        if (
            self.force_create or
            not param.optional or
            tuple(filter(None, kwargs.values()))
        ):
            ctx.params[param.name] = self._decorated(**kwargs)
        else:
            ctx.params[param.name] = None


COMPOSITES: Dict[Type, Composite] = {}


def has_composite(type_: Type):
    return type_ in COMPOSITES


def register_composite(type_: Type, composite: Composite):
    # Todo: warn about overwriting existing composite
    COMPOSITES[type_] = composite


def get_composite(type_: Type) -> Optional[Composite]:
    comp = COMPOSITES.get(type_)
    if comp is None and hasattr(type_, "__origin__"):
        comp = COMPOSITES.get(type_.__origin__)
    return comp


# noinspection PyPep8Naming
class composite_type(Composite[type]):
    """

    """
    @property
    def _match_type(self):
        return self._decorated


# noinspection PyPep8Naming
class composite_factory(Composite[Callable]):
    """
    Annotates a function that returns an instance of a composite type.

    Args:
        dest_type: The composite type, i.e. the type that will be recognized in the
            signature of the command function and matched with this factory function.
            If not specified, it is inferred from the return type.
        ignore_type_args: Whether type args can safely be ignored when matching.
        keep_underscores: Whether underscores should be retained in option names
            (True) or converted to hyphens (False).
        short_names: Dictionary mapping parameter names to short names. If
            not specified, usage of short names depends on `infer_short_names`.
            Set a value to `None` to disable short name usage for a paramter.
        infer_short_names: Whether to infer short names from parameter names.
            See Details on the algorithm used to select the short name. If a
            parameter has a short name specified in `short_names` it overrides
            the inferred short name.
        option_order: Specify an order of option processing that is different
            from the order in the signature of the annotated function.
        types: Dict mapping parameter names to functions that perform type
            conversion. By default, the type of a parameter is inferred from
            its annotation.
        positionals_as_options: Whether to treat positional arguments as required
            options.
        conditionals: Dict mapping paramter names or tuples of parameter names to
            conditional functions or lists of conditinal functions.
        validations: Dict mapping paramter names or tuples of parameter names to
            validation functions or lists of validation functions.
        required: Sequence of required options. If not specified, only paramters
            without default values are required.
        show_defaults: Whether to show defaults in the help text.
        hidden: Sequence of hidden options. These options are not displayed in the
            help text.
        param_help: Dict mapping parameters to help strings. By default, these are
            extracted from the function docstring.
    """
    def __init__(
        self,
        dest_type: Optional[Type] = None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self._target = dest_type

    @property
    def _match_type(self):
        return self._target

    def _create_decorator(self) -> Callable:
        if self._target is None:
            self._target = get_dest_type(self._decorated)
        return super()._create_decorator()


def create_composite(to_wrap: Union[Callable, Type], **kwargs) -> Composite:
    """Creates a :class:`Composite` for use in the `composites` paramter to a
    `command` or `group` decorator.

    Kwargs:
        to_wrap: The function/class to wrap.
        keep_underscores: Whether underscores should be retained in option names
            (True) or converted to hyphens (False).
        short_names: Dictionary mapping parameter names to short names. If
            not specified, usage of short names depends on `infer_short_names`.
            Set a value to `None` to disable short name usage for a paramter.
        infer_short_names: Whether to infer short names from parameter names.
            See Details on the algorithm used to select the short name. If a
            parameter has a short name specified in `short_names` it overrides
            the inferred short name.
        option_order: Specify an order of option processing that is different
            from the order in the signature of the annotated function.
        types: Dict mapping parameter names to functions that perform type
            conversion. By default, the type of a parameter is inferred from
            its annotation.
        positionals_as_options: Whether to treat positional arguments as required
            options.
        conditionals: Dict mapping paramter names or tuples of parameter names to
            conditional functions or lists of conditinal functions.
        validations: Dict mapping paramter names or tuples of parameter names to
            validation functions or lists of validation functions.
        required: Sequence of required options. If not specified, only paramters
            without default values are required.
        show_defaults: Whether to show defaults in the help text.
        hidden: Sequence of hidden options. These options are not displayed in the
            help text.
        param_help: Dict mapping parameters to help strings. By default, these are
            extracted from the function docstring.
    """
    if inspect.isclass(to_wrap):
        comp = composite_type(**kwargs)
    else:
        comp = composite_factory(**kwargs)
    comp(to_wrap)
    return comp



import pkg_resources

from autoclick.utils import LOG, set_global
from autoclick.commands import command, group
from autoclick.core import SignatureError, ParameterCollisionError
from autoclick.composites import (
    TypeCollisionError,
    create_composite,
    composite_type,
    composite_factory
)
from autoclick.composites.library import *
from autoclick.types import conversion
from autoclick.types.library import *
from autoclick.validations import ValidationError, validation
from autoclick.validations.library import *


# Load modules for validation and composite plugins
for entry_point in pkg_resources.iter_entry_points(group='autoclick'):
    LOG.debug("Loading plugin entry-point %s", str(entry_point))
    entry_point.load()



from abc import ABCMeta, abstractmethod
from typing import (
    Any, Callable, Dict, Iterable, Optional, Sequence, Set, Tuple, Type, Union, cast
)

import click

from autoclick.composites import Composite, get_composite
from autoclick.core import DEC, BaseDecorator, ParameterInfo, apply_to_parsed_args
from autoclick.types import Aggregate, AggregateTypeMixin
from autoclick.utils import EMPTY, LOG, get_global


class CommandMixin:
    """
    Mixin class that overrides :func:`parse_args` to apply validations and conditionals,
    and to resolve composite types.
    """
    def __init__(
        self,
        *args,
        conditionals: Dict[Sequence[str], Sequence[Callable]],
        validations: Dict[Sequence[str], Sequence[Callable]],
        composite_callbacks: Sequence[Callable[[dict], None]],
        aggregate_callbacks: Sequence[Callable[[dict], None]],
        used_short_names: Set[str],
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._conditionals = conditionals or {}
        self._validations = validations or {}
        self._composite_callbacks = composite_callbacks or {}
        self._aggregate_callbacks = aggregate_callbacks or {}
        self._used_short_names = used_short_names or {}

    def parse_args(self, ctx, args):
        click.Command.parse_args(cast(click.Command, self), ctx, args)
        apply_to_parsed_args(self._conditionals, ctx.params, update=True)
        apply_to_parsed_args(self._validations, ctx.params, update=False)
        for callback in self._composite_callbacks:
            callback(ctx)
        for callback in self._aggregate_callbacks:
            callback(ctx)
        return args

    @staticmethod
    def parse_extra_kwargs(extra_kwargs: Iterable[str]) -> dict:
        def parse_kwarg(kwarg) -> Tuple[str, str]:
            k, v = kwarg.split("=")
            k = k.lstrip("-")
            return k, v
        return dict(parse_kwarg(kwarg) for kwarg in extra_kwargs)


class AutoClickCommand(CommandMixin, click.Command):
    """
    Subclass of :class:`click.Command` that also inherits :class:`CommandMixin`.
    """
    pass


class AutoClickGroup(CommandMixin, click.Group):
    """
    Subclass of :class:`click.Group` that also inherits :class:`CommandMixin`.

    Args:
        match_prefix: Whether to look for a command that starts with the specified
            name if the command name cannot be matched exactly.
    """

    def __init__(self, *args, match_prefix: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self._match_prefix = match_prefix

    def command(
        self,
        name: Optional[str] = None,
        decorated: Optional[Callable] = None,
        **kwargs
    ):
        """A shortcut decorator for declaring and attaching a command to
        the group.  This takes the same arguments as :func:`command` but
        immediately registers the created command with this instance by
        calling into :meth:`add_command`.
        """
        def decorator(f):
            cmd = command(
                name=name,
                used_short_names=self._used_short_names,
                **kwargs
            )
            click_command = cmd(f)
            self.add_command(click_command)
            return click_command

        if decorated:
            return decorator(decorated)
        else:
            return decorator

    def group(
        self,
        name: Optional[str] = None,
        decorated: Optional[Callable] = None,
        **kwargs
    ):
        """A shortcut decorator for declaring and attaching a group to
        the group.  This takes the same arguments as :func:`group` but
        immediately registers the created command with this instance by
        calling into :meth:`add_command`.
        """
        def decorator(f):
            grp = group(
                name=name,
                used_short_names=self._used_short_names,
                **kwargs
            )
            click_group = grp(f)
            self.add_command(click_group)
            return click_group

        if decorated:
            return decorator(decorated)
        else:
            return decorator

    def get_command(self, ctx, cmd_name):
        cmd = click.Group.get_command(self, ctx, cmd_name)
        if cmd is not None:
            return cmd

        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        else:
            ctx.fail(f"Too many matches: {', '.join(sorted(matches))}")

    def parse_args(self, ctx, args):
        if not args and self.no_args_is_help and not ctx.resilient_parsing:
            click.echo(ctx.get_help(), color=ctx.color)
            ctx.exit()

        rest = CommandMixin.parse_args(self, ctx, args)
        if self.chain:
            ctx.protected_args = rest
            ctx.args = []
        elif rest:
            ctx.protected_args, ctx.args = rest[:1], rest[1:]

        return ctx.args


class DefaultAutoClickGroup(AutoClickGroup):
    """

    """
    def __init__(
        self,
        *args,
        invoke_without_command: bool = None,
        no_args_is_help: bool = None,
        default: Optional[str] = None,
        default_if_no_args: bool = False,
        **kwargs
    ):
        if default_if_no_args:
            if invoke_without_command is False or no_args_is_help is True:
                raise ValueError(
                    "One one of 'no_args_is_help', 'default_if_no_args' may be True."
                )
            invoke_without_command = True
            no_args_is_help = False

        super().__init__(
            *args, invoke_without_command=invoke_without_command,
            no_args_is_help=no_args_is_help, **kwargs
        )
        self._default_cmd_name = default
        self._default_if_no_args = default_if_no_args

    def set_default_command(self, cmd):
        """Sets a command function as the default command.
        """
        self._default_cmd_name = cmd.name
        self.add_command(cmd)

    def parse_args(self, ctx, args):
        if not args and self._default_if_no_args:
            args.insert(0, self._default_cmd_name)
        return super().parse_args(ctx, args)

    def get_command(self, ctx, cmd_name):
        if cmd_name not in self.commands:
            # No command name matched.
            ctx.arg0 = cmd_name
            cmd_name = self._default_cmd_name
        return super().get_command(ctx, cmd_name)

    def resolve_command(self, ctx, args):
        base = super()
        cmd_name, cmd, args = base.resolve_command(ctx, args)
        if hasattr(ctx, 'arg0'):
            args.insert(0, ctx.arg0)
        return cmd_name, cmd, args

    def format_commands(self, ctx, formatter):
        formatter = DefaultCommandFormatter(self, formatter, mark='*')
        return super().format_commands(ctx, formatter)


class DefaultCommandFormatter:
    """Wraps a formatter to mark a default command.
    """

    def __init__(self, group_, formatter, mark='*'):
        self._group = group_
        self._formatter = formatter
        self._mark = mark

    def __getattr__(self, attr):
        return getattr(self.formatter, attr)

    def write_dl(self, rows, *args, **kwargs):
        rows_ = []
        for cmd_name, help_str in rows:
            if cmd_name == self._group.default_cmd_name:
                rows_.insert(0, (cmd_name + self._mark, help_str))
            else:
                rows_.append((cmd_name, help))
        return self._formatter.write_dl(rows_, *args, **kwargs)


class BaseCommandDecorator(BaseDecorator[DEC], metaclass=ABCMeta):
    """
    Base class for decorators that wrap command functions.
    """
    def __init__(
        self,
        name: Optional[str] = None,
        composite_types: Optional[Dict[str, Composite]] = None,
        add_composite_prefixes: bool = True,
        command_help: Optional[str] = None,
        option_class: Type[click.Option] = click.Option,
        argument_class: Type[click.Argument] = click.Argument,
        used_short_names: Optional[Set[str]] = None,
        default_values: Optional[Dict[str, Any]] = None,
        version: Optional[Union[str, bool]] = None,
        pass_context: Optional[bool] = None,
        main: bool = False,
        **kwargs
    ):
        super().__init__(**kwargs)
        self._name = name
        self._composite_types: Dict[str, Composite] = composite_types or {}
        self._add_composite_prefixes = get_global(
            "add_composite_prefixes", add_composite_prefixes
        )
        self._command_help = command_help
        self._option_class = option_class
        self._argument_class = argument_class
        self._used_short_names = set()
        if used_short_names:
            self._used_short_names.update(used_short_names)
        self._default_values = default_values or {}
        self._pass_context = get_global("pass_context", pass_context)
        self._main = main
        self._add_version_option = version

    @property
    def name(self) -> str:
        """The command name."""
        return self._name or self._decorated.__name__.lower().replace('_', '-')

    def _handle_parameter_info(self, param: ParameterInfo) -> bool:
        if param.extra_arguments:
            LOG.warning("*args are not currently supported")
        elif param.extra_kwargs:
            LOG.warning("**kwargs are not currently supported")
        return super()._handle_parameter_info(param)

    def _create_decorator(self) -> DEC:
        parameter_infos = self._get_parameter_info()
        command_params = []
        composite_callbacks = []
        aggregate_callbacks = []

        # TODO
        # if self._add_version_option:
        #     command_params.append()

        if self._pass_context:
            ctx_param = list(parameter_infos.keys())[0]
            if parameter_infos[ctx_param].anno_type in {click.Context, EMPTY, None}:
                parameter_infos.pop(ctx_param)
                if ctx_param in self._option_order:
                    self._option_order.remove(ctx_param)
            else:
                LOG.warning(
                    "pass_context set to True, but first parameter of function %s "
                    "does not appear to be of type click.Context",
                    self.name
                )

        for param_name in self._option_order:
            param = parameter_infos[param_name]

            composite: Composite = (
                self._composite_types[param_name]
                if param_name in self._composite_types
                else get_composite(param.match_type)
            )

            if composite:
                click_parameters, callback = composite.create_click_parameters(
                    param=param,
                    used_short_names=self._used_short_names,
                    default_values=self._default_values,
                    add_prefixes=self._add_composite_prefixes,
                    hidden=param.name in self._hidden,
                    option_class=self._option_class,
                    argument_class=self._argument_class,
                    help_text=self._get_help(param.name),
                    force_positionals_as_options=self._positionals_as_options
                )
                command_params.extend(click_parameters)
                composite_callbacks.append(callback)
            else:
                command_params.append(self._create_click_parameter(
                    param=param,
                    used_short_names=self._used_short_names,
                    default_values=self._default_values,
                    option_class=self._option_class,
                    argument_class=self._argument_class
                ))

            if isinstance(param.click_type, AggregateTypeMixin):
                aggregate_callbacks.append(Aggregate(param.name, param.click_type))

        desc = None
        if self._docs and self._docs.description:
            desc = str(self._docs.description)

        callback = self._decorated
        if self._pass_context:
            callback = click.pass_context(callback)

        # Todo: this is pending release of click 7.1
        # if (
        #     "no_args_is_help" not in self._extra_kwargs and
        #     any(p.required for p in command_params)
        # ):
        #     # Pass `no_args_is_help=True` unless there are no required parameters
        #     self._extra_kwargs["no_args_is_help"] = True

        click_command = self._create_click_command(
            name=self.name,
            callback=callback,
            params=command_params,
            help=desc,
            conditionals=self._conditionals,
            validations=self._validations,
            composite_callbacks=composite_callbacks,
            aggregate_callbacks=aggregate_callbacks,
            **self._extra_kwargs
        )

        if (
            self._main and
            hasattr(self._decorated, "__module__") and
            self._decorated.__module__ == "__main__"
        ):
            import atexit
            atexit.register(click_command)

        return click_command

    @abstractmethod
    def _create_click_command(self, **kwargs) -> DEC:
        pass


# noinspection PyPep8Naming
class command(BaseCommandDecorator[click.Command]):
    """
    Decorator that creates a click.Command based on type annotations of the
    annotated function.

    Args:
        command_class: Class to use when creating the :class:`click.Command`. This must
            inherit from :class:`CommandMixin`.
        name: The command name. If not specified, it is taken from the name
            of the annotated function.
        composite_types: Dict mapping parameter names to :class:`CompositeParameter`
            objects.
        add_composite_prefixes: By default, the parameter name is added as a prefix
            when deriving the option names for composite parameters. If set to false,
            each composite type may only be used for at most one parameter, and the
            user must ensure that no composite parameter names conflict with each
            other or with other parameter names in the annotated function.
        default_values: Specify default values for parameters. The primary usage is
            to specify default values for hidden parameters of composite types.
            Otherwise, it is better to specify default values in the signature of the
            command function.
        command_help: Command description. By default, this is extracted from the
            funciton docstring.
        option_class: Class to use when creating :class:`click.Option`s.
        argument_class: Class to use when creating :class:`click.Argument`s.
        keep_underscores: Whether underscores should be retained in option names
            (True) or converted to hyphens (False).
        short_names: Dictionary mapping parameter names to short names. If
            not specified, usage of short names depends on `infer_short_names`.
            Set a value to `None` to disable short name usage for a paramter.
        infer_short_names: Whether to infer short names from parameter names.
            See Details on the algorithm used to select the short name. If a
            parameter has a short name specified in `short_names` it overrides
            the inferred short name.
        option_order: Specify an order of option processing that is different
            from the order in the signature of the annotated function.
        types: Dict mapping parameter names to functions that perform type
            conversion. By default, the type of a parameter is inferred from
            its annotation.
        positionals_as_options: Whether to treat positional arguments as required
            options.
        conditionals: Dict mapping paramter names or tuples of parameter names to
            conditional functions or lists of conditinal functions.
        validations: Dict mapping paramter names or tuples of parameter names to
            validation functions or lists of validation functions.
        required: Sequence of required options. If not specified, only paramters
            without default values are required.
        show_defaults: Whether to show defaults in the help text.
        hidden: Sequence of hidden options. These options are not displayed in the
            help text.
        param_help: Dict mapping parameters to help strings. By default, these are
            extracted from the function docstring.
        pass_context: Whether to pass in the click.Context as the first argument
            to the function.
        kwargs: Any extra keyword arguments are passed to the Command.
    """
    def __init__(
        self,
        name: str = None,
        command_class: Type[CommandMixin] = AutoClickCommand,
        **kwargs
    ):
        super().__init__(name=name, **kwargs)
        self._command_class = command_class

    def _create_click_command(self, **kwargs) -> click.Command:
        return cast(click.Command, self._command_class(
            used_short_names=self._used_short_names,
            **kwargs
        ))


# noinspection PyPep8Naming
class group(BaseCommandDecorator[click.Group]):
    """
    Decorator that creates a :class:`click.Group` based on type annotations of the
    annotated function.

    Args:
        group_class: Class to use when creating the :class:`click.Group`. This must
            inherit from :class:`CommandMixin`.
        name: The command name. If not specified, it is taken from the name
            of the annotated function.
        composite_types: Dict mapping parameter names to :class:`CompositeParameter`
            objects.
        add_composite_prefixes: By default, the parameter name is added as a prefix
            when deriving the option names for composite parameters. If set to false,
            each composite type may only be used for at most one parameter, and the
            user must ensure that no composite parameter names conflict with each
            other or with other parameter names in the annotated function.
        default_values: Specify default values for parameters. The primary usage is
            to specify default values for hidden parameters of composite types.
            Otherwise, it is better to specify default values in the signature of the
            command function.
        command_help: Command description. By default, this is extracted from the
            funciton docstring.
        option_class: Class to use when creating :class:`click.Option`s.
        argument_class: Class to use when creating :class:`click.Argument`s.
        keep_underscores: Whether underscores should be retained in option names
            (True) or converted to hyphens (False).
        short_names: Dictionary mapping parameter names to short names. If
            not specified, usage of short names depends on `infer_short_names`.
            Set a value to `None` to disable short name usage for a paramter.
        infer_short_names: Whether to infer short names from parameter names.
            See Details on the algorithm used to select the short name. If a
            parameter has a short name specified in `short_names` it overrides
            the inferred short name.
        option_order: Specify an order of option processing that is different
            from the order in the signature of the annotated function.
        types: Dict mapping parameter names to functions that perform type
            conversion. By default, the type of a parameter is inferred from
            its annotation.
        positionals_as_options: Whether to treat positional arguments as required
            options.
        conditionals: Dict mapping paramter names or tuples of parameter names to
            conditional functions or lists of conditinal functions.
        validations: Dict mapping paramter names or tuples of parameter names to
            validation functions or lists of validation functions.
        required: Sequence of required options. If not specified, only paramters
            without default values are required.
        show_defaults: Whether to show defaults in the help text.
        hidden: Sequence of hidden options. These options are not displayed in the
            help text.
        param_help: Dict mapping parameters to help strings. By default, these are
            extracted from the function docstring.
        pass_context: Whether to pass in the click.Context as the first argument
            to the function.
        kwargs: Any extra kwargs are passed to the Group.
    """
    def __init__(
        self,
        name: str = None,
        group_class: Type[CommandMixin] = AutoClickGroup,
        commands: Optional[Dict[str, click.Command]] = None,
        **kwargs
    ):
        super().__init__(name=name, **kwargs)
        self._group_class = group_class
        self._extra_kwargs["commands"] = commands or {}

    def _create_click_command(self, **kwargs) -> click.Group:
        return cast(click.Group, self._group_class(
            used_short_names=self._used_short_names,
            **kwargs
        ))



import logging
from pathlib import Path
import sys
from typing import Optional

from autoclick.composites import composite_factory


@composite_factory(hidden=["log_name"])
def log(
    log_name: str = "DEFAULT", log_level: str = "WARN", log_file: Optional[Path] = None
) -> logging.Logger:
    """

    Args:
        log_name:
        log_level:
        log_file:

    Returns:

    """
    logger = logging.getLogger(log_name)
    logger.setLevel(log_level)
    if log_file:
        logger.addHandler(logging.FileHandler(log_file))
    else:
        logger.addHandler(logging.StreamHandler(sys.stderr))
    return logger



from abc import ABCMeta, abstractmethod
import inspect
import re
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Generic,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast
)

import click
import docparse

from autoclick.types import get_conversion
from autoclick.types.library import OptionalTuple
from autoclick.utils import EMPTY, EMPTY_OR_NONE, LOG, get_global
from autoclick.validations import get_validations


UNDERSCORES = re.compile("_")
ALPHA_CHARS = set(chr(i) for i in tuple(range(97, 123)) + tuple(range(65, 91)))
DEC = TypeVar("DEC")


class ParameterCollisionError(Exception):
    """Raised when a composite paramter has the same name as one in the parent
    function.
    """


class SignatureError(Exception):
    """Raised when the signature of the decorated method is not supported.
    """


class ParameterInfo:
    """Extracts and contains the necessary information from a
    :class:`inspect.Parameter`.

    Args:
        name: The parameter name.
        param: A :class:`inspect.Parameter`.
        click_type: The conversion function, if specified explicitly.
        required: Whether this is explicitly specified to be a required parameter.
    """

    def __init__(
        self, name: str, param: inspect.Parameter, click_type: Optional[type] = None,
        required: bool = False
    ):
        self.name = name
        self.anno_type = param.annotation
        self.anno_args = None
        self.click_type = click_type
        self.optional = not (required or param.default is EMPTY)
        self.default = None if param.default is EMPTY else param.default
        self.nargs = 1
        self.multiple = False
        self.extra_arguments = (param.kind is inspect.Parameter.VAR_POSITIONAL)
        self.extra_kwargs = (param.kind is inspect.Parameter.VAR_KEYWORD)

        if self.anno_type in EMPTY_OR_NONE:
            if not self.optional:
                LOG.debug(
                    f"No type annotation or default value for parameter "
                    f"{name}; using <str>"
                )
                self.anno_type = str
            else:
                self.anno_type = type(self.default)
                LOG.debug(
                    f"Inferring type {self.anno_type} from paramter {name} "
                    f"default value {self.default}"
                )
        elif isinstance(self.anno_type, str):
            if self.anno_type in globals():
                self.anno_type = globals()[self.anno_type]
            else:
                raise SignatureError(
                    f"Could not resolve type {self.anno_type} of paramter {name}"
                )

        # Resolve Union attributes
        # The only time a Union type is allowed is when it has two args and
        # one is None (i.e. an Optional)
        if (
            hasattr(self.anno_type, "__origin__") and
            self.anno_type.__origin__ is Union
        ):
            filtered_args = set(self.anno_type.__args__)
            if type(None) in filtered_args:
                filtered_args.remove(type(None))
            if len(filtered_args) == 1:
                self.anno_type = filtered_args.pop()
                self.optional = True
                self.default = None
            else:
                raise SignatureError(
                    f"Union type not supported for parameter {name}"
                )

        self.match_type = self.anno_type

        def resolve_new_type(t):
            return t.__supertype__ if (
                inspect.isfunction(t) and hasattr(t, "__supertype__")
            ) else t

        self.anno_type = resolve_new_type(self.anno_type)

        # Resolve meta-types
        if hasattr(self.anno_type, "__origin__"):
            origin = self.anno_type.__origin__

            if hasattr(self.anno_type, "__args__"):
                if origin == Tuple:
                    # Resolve Tuples with specified arguments
                    if self.click_type is None:
                        self.click_type = click.Tuple([
                            resolve_new_type(a) for a in self.anno_type.__args__
                        ])
                elif len(self.anno_type.__args__) == 1:
                    self.match_type = self.anno_type.__args__[0]
                else:
                    self.anno_args = self.anno_type.__args__

            self.anno_type = origin

        # Unwrap complex types
        while hasattr(self.anno_type, "__extra__"):
            self.anno_type = self.anno_type.__extra__

        # Set nargs when type is a click.Tuple
        if isinstance(self.click_type, click.Tuple):
            self.nargs = len(cast(click.Tuple, self.click_type).types)
            if self.default is None:
                # Substitute a subclass of click.Tuple that will convert a sequence
                # of all None's to a None
                self.default = (None,) * self.nargs
                self.click_type = OptionalTuple(self.click_type.types)
            elif not isinstance(self.default, Sequence):
                raise SignatureError(
                    f"Default value of paramter {self.name} of type Tuple must be a "
                    f"sequence."
                )
            else:
                arrity = len(tuple(self.default))
                if arrity != self.nargs:
                    raise SignatureError(
                        f"Default value of paramter {self.name} of type Tuple must be "
                        f"a collection having the same arrity; {arrity} != {self.nargs}"
                    )

        # Collection types are treated as parameters that can be specified
        # multiple times
        if (
            self.nargs == 1 and
            self.anno_type != str and
            issubclass(self.anno_type, Collection)
        ):
            self.multiple = True

        if self.match_type is None:
            self.match_type = self.anno_type

        if self.click_type is None:
            self.click_type = get_conversion(
                self.match_type, self.anno_type, self.anno_args
            )

        self.is_flag = (
            self.click_type == bool or
            isinstance(self.click_type, click.types.BoolParamType)
        )


class BaseDecorator(Generic[DEC], metaclass=ABCMeta):
    """
    Base class for decorators of groups, commands, and composites.
    """

    def __init__(
        self,
        keep_underscores: bool = False,
        short_names: Optional[Dict[str, str]] = None,
        infer_short_names: bool = True,
        option_order: Optional[Sequence[str]] = None,
        types: Optional[Dict[str, Callable]] = None,
        positionals_as_options: bool = False,
        conditionals: Dict[
            Union[str, Tuple[str, ...]], Union[Callable, List[Callable]]] = None,
        validations: Dict[
            Union[str, Tuple[str, ...]], Union[Callable, List[Callable]]] = None,
        required: Optional[Sequence[str]] = None,
        hidden: Optional[Sequence[str]] = None,
        show_defaults: bool = False,
        param_help: Optional[Dict[str, str]] = None,
        decorated: Optional[Callable] = None,
        **extra_kwargs
    ):
        self._keep_underscores = get_global("keep_underscores", keep_underscores)
        self._short_names = short_names or {}
        self._infer_short_names = get_global("infer_short_names", infer_short_names)
        self._option_order = option_order or []
        self._positionals_as_options = positionals_as_options
        self._types = types or {}
        self._required = required or set()
        self._hidden = hidden or set()
        self._show_defaults = show_defaults
        self._param_help = param_help or {}
        self._decorated = None
        self._docs = None
        self._extra_kwargs = extra_kwargs

        def _as_many_to_many(d):
            if d is None:
                return {}
            else:
                return dict(
                    (
                        k if isinstance(k, tuple) else (k,),
                        [v] if v and not isinstance(v, list) else v
                    )
                    for k, v in d.items()
                )

        self._conditionals = _as_many_to_many(conditionals)
        self._validations = _as_many_to_many(validations)

        if decorated:
            self(decorated=decorated)

    def __call__(self, decorated: Callable) -> DEC:
        self._decorated = decorated
        # TODO: support other docstring styles
        self._docs = docparse.parse_docs(decorated, docparse.DocStyle.GOOGLE)
        return self._create_decorator()

    @abstractmethod
    def _create_decorator(self) -> DEC:
        pass

    def _get_parameter_info(self) -> Dict[str, ParameterInfo]:
        if inspect.isclass(self._decorated):
            signature_parameters = dict(
                inspect.signature(cast(type, self._decorated).__init__).parameters
            )
            signature_parameters.pop("self")
        else:
            signature_parameters = dict(
                inspect.signature(cast(Callable, self._decorated)).parameters
            )

        parameter_infos = {}

        for name, sig_param in signature_parameters.items():
            param = ParameterInfo(
                name, sig_param, self._types.get(name, None), name in self._required
            )
            if self._handle_parameter_info(param):
                parameter_infos[name] = param

        return parameter_infos

    def _handle_parameter_info(self, param: ParameterInfo) -> bool:
        """
        Register parameter. Subclasses can override this method to filter out
        some paramters.

        Args:
            param: A :class:`ParameterInfo`.

        Returns:
            True if this parameter should be added to the parser.
        """

        if param.name not in self._option_order:
            self._option_order.append(param.name)

        vals = get_validations(param.match_type)
        if vals:
            if param.name not in self._validations:
                self._validations[(param.name,)] = []
            self._validations[(param.name,)].extend(vals)

        return True

    def _create_click_parameter(
        self,
        param: ParameterInfo,
        used_short_names: Set[str],
        default_values: Dict[str, Any],
        option_class: Type[click.Option],
        argument_class: Type[click.Argument],
        long_name_prefix: Optional[str] = None,
        hidden: bool = False,
        force_positionals_as_options: bool = False
    ) -> click.Parameter:
        """Create a click.Parameter instance (either Option or Argument).

        Args:
            param: A :class:`ParameterInfo`.
            used_short_names: A set of short names that have been used by other
                parameters and thus should not be re-used.
            default_values:
            option_class: Class to instantiate for option parameters.
            argument_class: Class to instantiate for argument parameters.
            long_name_prefix: Prefix to add to long option names.
            hidden: Whether to not show the parameter in help text.
            force_positionals_as_options: Whether to force positional arguments to be
                treated as options.

        Returns:
            A :class:`click.Parameter`.
        """

        param_name = param.name
        long_name = self._get_long_name(param_name, long_name_prefix)

        if (
            param.optional or
            force_positionals_as_options or
            self._positionals_as_options
        ):
            if not param.is_flag:
                long_name_decl = f"--{long_name}"
            elif long_name.startswith("no-"):
                long_name_decl = f"--{long_name[3:]}/--{long_name}"
            else:
                long_name_decl = f"--{long_name}/--no-{long_name}"

            param_decls = [long_name_decl]

            short_name = self._get_short_name(param_name, used_short_names)
            if short_name:
                used_short_names.add(short_name)
                param_decls.append(f"-{short_name}")

            return option_class(
                param_decls,
                type=None if param.is_flag else param.click_type,
                required=not param.optional,
                default=default_values.get(param_name, param.default),
                show_default=self._show_defaults,
                nargs=param.nargs,
                hidden=hidden or param_name in self._hidden,
                multiple=param.multiple,
                help=self._get_help(param_name)
            )
        else:
            # TODO: where to show argument help?
            return argument_class(
                [long_name],
                type=param.click_type,
                default=default_values.get(param_name, param.default),
                nargs=-1 if param.nargs == 1 and param.multiple else param.nargs
            )

    def _get_short_name(self, name: str, used_short_names: Set[str]):
        short_name = self._short_names.get(name, None)

        if short_name and short_name in used_short_names:
            raise ParameterCollisionError(
                f"Short name {short_name} defined for two different parameters"
            )
        elif not short_name and self._infer_short_names:
            for char in name:
                if char.isalpha():
                    if char.lower() not in used_short_names:
                        short_name = char.lower()
                    elif char.upper() not in used_short_names:
                        short_name = char.upper()
                    else:
                        continue
                    break
            else:
                # try to select one randomly
                remaining = ALPHA_CHARS - used_short_names
                if len(remaining) == 0:
                    raise click.BadParameter(
                        f"Could not infer short name for parameter {name}"
                    )
                # TODO: this may not be deterministic
                short_name = remaining.pop()

        return short_name

    def _get_long_name(self, name: str, prefix: Optional[str] = None):
        long_name = name
        if prefix:
            long_name = f"{prefix}_{long_name}"
        if not self._keep_underscores:
            long_name = UNDERSCORES.sub("-", long_name)
        return long_name

    def _get_help(self, name: str):
        if name in self._param_help:
            return self._param_help[name]
        elif self._docs and self._docs.parameters and name in self._docs.parameters:
            return str(self._docs.parameters[name].description)


def apply_to_parsed_args(d, values: dict, update=False):
    for params, fns in d.items():
        fn_kwargs = dict(
            (param, values.get(param, None))
            for param in params
        )
        for fn in fns:
            result = fn(**fn_kwargs)
            if result and update:
                for param, value in result.items():
                    values[param] = value



import inspect
import logging
from typing import Callable, Optional, Type, TypeVar


LOG = logging.getLogger("AutoClick")
EMPTY = inspect.Signature.empty
EMPTY_OR_NONE = {EMPTY, None}
GLOBAL_CONFIG = {}
T = TypeVar("T")


def set_global(name: str, value: T) -> Optional[T]:
    """
    Configure global AutoClick settings:

    * "infer_short_names": (bool) whether to always/never infer paramter short names.
         (defaut=True)
    * "keep_underscores": (bool) whether to retain underscores in paramter long names
        or covert them to dashes. (defaut=False)
    * "pass_context": (bool) whether to always/never pass the context to command
        functions, so that it is not required to specify pass_context=True to every
        command/group decorator. (defaut=False)

    Args:
        name: The global parameter name.
        value: The paramter value.

    Returns:
        The previous value of the setting.
    """
    prev = GLOBAL_CONFIG.get(name, None)
    if prev != value:
        GLOBAL_CONFIG[name] = value
    return prev


def get_global(name: str, default: T) -> T:
    return GLOBAL_CONFIG.get(name, default)


def get_match_type(f: Callable) -> Type:
    params = inspect.signature(f).parameters
    if len(params) == 0:
        raise ValueError(f"Function {f} must have at least one parameter")
    params = list(params.values())
    if len(params) > 1:
        for p in params[1:]:
            if p.default == EMPTY:
                raise ValueError(
                    f"All but the first parameter must have default values in "
                    f"the signature of function {f}."
                )
    return params[0].annotation


def get_dest_type(f: Callable) -> Type:
    dest_type = inspect.signature(f).return_annotation
    if dest_type in EMPTY_OR_NONE:
        raise ValueError(f"Function {f} must have a non-None return annotation")
    return dest_type



import re
from typing import List, Sequence

from autoclick.types import *
from autoclick.utils import get_match_type


VALIDATIONS: Dict[Type, List[Callable]] = {}

UNDERSCORES = re.compile("_")
ALPHA_CHARS = set(chr(i) for i in tuple(range(97, 123)) + tuple(range(65, 91)))


class ValidationError(click.UsageError):
    """Raised by a validation function when an input violates a constraint.
    """


def register_validation(type_: Type, validation_fn: Callable):
    """

    Args:
        type_:
        validation_fn:
    """
    if type_ not in VALIDATIONS:
        VALIDATIONS[type_] = []
    VALIDATIONS[type_].append(validation_fn)


def has_validations(type_: Type) -> bool:
    """

    Args:
        type_:

    Returns:

    """
    return type_ in VALIDATIONS


def get_validations(type_: Type) -> Optional[Sequence[Callable]]:
    """

    Args:
        type_:

    Returns:

    """
    return VALIDATIONS.get(type_, None)


def validation(
    match_type: Optional[Type] = None,
    depends: Optional[Tuple[Callable, ...]] = None,
    decorated: Optional[Callable] = None
):
    """Annotates a single-parameter validation.

    Args:
        match_type: The type that will match this validation. If None, is inferred
            from the type of the first parameter in the signature of the annotated
            function.
        depends: Other validations that are pre-requisite for this one.
        decorated: The function to decorate.

    Returns:
        A decorator function.
    """
    def decorator(f: Callable) -> Callable:
        _match_type = match_type
        if _match_type is None:
            _match_type = get_match_type(f)

        if depends:
            def composite_validation(**kwargs):
                for dep in depends:
                    dep(**kwargs)
                f(**kwargs)
            target = composite_validation
        else:
            target = f

        # Annotated validation functions can only ever validate a single parameter
        # so we can explicitly specify the param name and value as kwargs to the
        # decorated function.
        def call_target(**kwargs):
            if len(kwargs) == 2 and set(kwargs.keys()) == {"param_name", "value"}:
                pass
            elif len(kwargs) != 1:
                print(kwargs)
                raise ValueError(
                    "A @validation decorator may only validate a single parameter."
                )
            else:
                kwargs = dict(zip(("param_name", "value"), list(kwargs.items())[0]))
            if kwargs["value"] is not None:
                target(**kwargs)

        register_validation(match_type, call_target)

        return call_target

    if decorated:
        return decorator(decorated)
    else:
        return decorator



from abc import ABCMeta, abstractmethod
from typing import Callable, Collection, Dict, List, Optional, Tuple, Type, Union, cast

import click

from autoclick.utils import get_dest_type


CONVERSIONS: Dict[Type, click.ParamType] = {}
AUTOCONVERSIONS = []
DEFAULT_METAVAR = "ARG"


class ParamTypeAdapter(click.ParamType):
    """
    Adapts a conversion function to a :class:`click.ParamType`.

    Args:
        name:
        fn:
    """
    def __init__(self, name, fn):
        self.name = name
        self.fn = fn

    def convert(self, value, param, ctx):
        return self.fn(value, param, ctx)


class AggregateTypeMixin(metaclass=ABCMeta):
    @abstractmethod
    def aggregate(self, values):
        pass


class Aggregate:
    def __init__(self, param_name: str, param_type: AggregateTypeMixin):
        self.param_name = param_name
        self.param_type = param_type

    def __call__(self, ctx: click.Context):
        if self.param_name in ctx.params:
            agg_value = self.param_type.aggregate(ctx.params.pop(self.param_name))
            ctx.params[self.param_name] = agg_value


def conversion(
    dest_type: Optional[Type] = None,
    depends: Optional[Tuple[Callable, ...]] = None,
    decorated: Optional[Callable] = None,
    name: Optional[str] = None
) -> Callable:
    """Annotates a conversion function.

    Args:
        dest_type: Destination type for this conversion. If None, it is
            inferred from the return type of the annotated function.
        depends: Functions on which this conversion depends. They are called in
            order, with the output from each function being passed as the input
            to the next. The type of the parameter to the conversion function
            must be the return type of the last dependency.
        decorated: The function to decorate.
        name: Type name; defaults to the `name` attribute of `dest_type`.

    Returns:
        A decorator function.
    """
    def decorator(f: Callable) -> Callable:
        _name = name
        _dest_type = dest_type

        if _dest_type is None:
            _dest_type = get_dest_type(f)

        if depends:
            def composite_conversion(value):
                for dep in depends:
                    value = dep(value)
                return f(value)

            target = composite_conversion
        else:
            target = f

        if _name is None:
            if hasattr(_dest_type, "__name__"):
                _name = _dest_type.__name__
            elif hasattr(_dest_type, "_name"):
                _name = _dest_type._name
            else:
                raise ValueError(
                    f"Name cannot be determined from {_dest_type}; the 'name' "
                    f"argument must be provided."
                )

        click_type = ParamTypeAdapter(_name, target)
        register_conversion(_dest_type, click_type)
        return target

    if decorated:
        return decorator(decorated)
    else:
        return decorator


def register_conversion(type_: Type, click_type: click.ParamType):
    """

    Args:
        type_:
        click_type:
    """
    CONVERSIONS[type_] = click_type


def has_conversion(type_: Type) -> bool:
    """

    Args:
        type_:

    Returns:

    """
    return type_ in CONVERSIONS


def get_conversion(
    match_type: Type, true_type: Optional[Type] = None,
    type_args: Optional[List[Type]] = None
) -> Callable:
    """
    Gets a conversion function for the given type, if one is available.

    Args:
        match_type: Type to match against.
        true_type: Type to auto-convert, if `match_type` does not match any
            registered conversions.
        type_args: If the original type was generic, this is the list of type arguments.

    Returns:
        A Callable - generally either a click.ParamType (if a conversion was found) or
        `true_type`.
    """
    if match_type in CONVERSIONS:
        return CONVERSIONS[match_type]

    if true_type is None:
        true_type = match_type

    for filter_fn, conversion_fn, pass_type in AUTOCONVERSIONS:
        if filter_fn(true_type):
            args = []
            if pass_type:
                args.append(true_type)
            if type_args:
                args.append(type_args)
            return conversion_fn(*args)

    if issubclass(true_type, Collection):
        return match_type

    return true_type


def register_autoconversion(
    filter_fn: Union[Type, Callable[[Type], bool]],
    conversion_fn: Callable[[Type, Optional[List[Type]]], click.ParamType],
    pass_type: bool = True
):
    if isinstance(filter_fn, type):
        filter_type = cast(type, filter_fn)
        filter_fn = lambda type_: issubclass(type_, filter_type)
    AUTOCONVERSIONS.append((filter_fn, conversion_fn, pass_type))


def autoconversion(
    filter_fn: Union[Type, Callable[[Type], bool]],
    conversion_fn: Optional[Callable[[Type], click.ParamType]] = None,
    pass_type: bool = True
):
    """
    Decorator that registers an automatic conversion for a (usually built-in) type.

    Args:
        filter_fn: Function that returns a boolean indicating whether or not
            the autoconversion applies to a given type.
        conversion_fn: Function that returns a `click.ParamType` for a given type.
        pass_type: Whether to pass the type being converted as the first argument
            to the conversion function (defaults to True).

    Returns:
        An object that is a subclass of `click.ParamType`.
    """
    def decorator(fn: Callable[[Type], click.ParamType]):
        register_autoconversion(filter_fn, conversion_fn or fn, pass_type)
        return fn
    return decorator



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['autoclick',
 'autoclick.composites',
 'autoclick.types',
 'autoclick.validations']

package_data = \
{'': ['*']}

install_requires = \
['click==7.0', 'docparse']

entry_points = \
{'autoclick': ['composites = autoclick.composites',
               'validations = autoclick.validations']}

setup_kwargs = {
    'name': 'autoclick',
    'version': '0.8.0',
    'description': 'Auto-generate Click-based CLIs from python3 type annotations.',
    'long_description': '# AutoClick\n\nAutoClick creates Click-based CLIs using type annotations.\n\nThe simplest use of AutoClick requires annotating your main method with `@autoclick.command`:\n\n```python\n# test.py\nimport autoclick\n\n@autoclick.command("greet")\ndef main(greeting: str, name: str):\n    print(f"{greeting} {name}")\n\nif __name__ == "__main__":\n    main()\n```\n\n```bash\n$ python test.py --help\nUsage: test.py [OPTIONS] [GREETING] [NAME]\n\nOptions:\n  --help  Show this message and exit.\n```\n\nFor additional customization, keyword arguments can be passed to the command decorator:\n\n```python\nimport autoclick\n\n@autoclick.command(\n    short_names={\n        "greeting": "G",\n        "name": "n"\n    },\n    show_defaults=True\n)\ndef main(greeting: str = "hello", name: str = "human"):\n    print(f"{greeting} {name}")\n```\n\n```bash\n$ python test.py --help\nUsage: test.py [OPTIONS]\n\nOptions:\n  -G, --greeting TEXT  [default: hello]\n  -n, --name TEXT      [default: human]\n  --help               Show this message and exit.\n```\n\n## Type conversion\n\nIn Click, type conversion can be done either in a callback or by using a callable type (such as a subclass of ParamType) as the type. In AutoClick, type conversions are performed automatically based on type annotations for callable types. However, for more complex type conversion, there are three additional methods:\n\n1. Automatic conversion functions. A conversion function is decorated by `@conversion`. The conversion decorator by default infers the type being converted to from the return annotation. Otherwise, the destination type can be specified as an argument to the decorator. The decorator registers the function as the converter for the specified type. When that type is encountered as an annotation of a parameter to a command function, the converter function is used to convert the string argument to that type.\n\n```python\nimport autoclick\n\nclass Bork:\n    def __init__(self, n: int):\n        self.n = n\n\n    def __str__(self):\n        print(",".join(["bork"] * self.n))\n\n@autoclick.conversion()\ndef bork(n: int) -> Bork:\n    return Bork(n)\n\n@autoclick.command("bork")\ndef main(b: Bork):\n    print(b)\n```\n\nIn the case where there needs to be specialized handling of common types, new types can be created using `typing.NewType`:\n\n```python\nimport autoclick\nimport typing\n\nDoubleInt = typing.NewType("DoubleInt", int)\n\n@autoclick.conversion(DoubleInt)\ndef double_int(i: str):\n    return int(i) * 2\n\n@autoclick.command("double")\ndef main(i1: int, i2: DoubleInt):\n    print(i1, i2)\n```\n\n2. Conversion functions can also be specified explicitly in the command decorator:\n\n```python\nimport autoclick\n\n@autoclick.command(\n    types={\n        "a": double_int\n    }\n)\ndef main(a: int):\n    print(a)\n```\n\nNote that any of the types in the `click.Types` package can also be used in this way.\n\n3. For composite types, the `autoclick.composite_type` and `autoclick.composite_factory` functions can be used. An example of a composite type is a class that requires more than one parameter to its constructor. For composite types, the same annotation-based CLI creation is performed, and the parameters are injected into the CLI in place of the composite parameter.\n\n```python\nimport autoclick\n@autoclick.composite_type()\nclass Foo:\n    def __init__(self, bar: str, baz: int):\n        self.bar = bar\n        self.baz = baz\n\n@autoclick.command()\ndef main(foo: Foo):\n    print(foo.bar, foo.baz)\n```\n\nIn this case, the options to main would be `--foo-bar` and `--foo-baz`. Once the CLI is processed, the values of these options are used to construct the `Foo` instance, which is then passed to the call to `main`. The parameter name in the command function is prepended to the parameter names of the composite type, so that a composite type can be used multiple types in a command function signature.\n\nA `autoclick.composite_factory` function returns a complex type. For example, the code below is equivalent to the code above:\n\n```python\nimport autoclick\n\nclass Foo:\n    def __init__(self, bar: str, baz: int):\n        ...\n\n@autoclick.composite_factory(Foo)\ndef foo_factory(bar: str, baz: int):\n    return Foo(bar, baz)\n```\n\n## Conditionals and Validations\n\nConditionals and Validations are similar - they are both decorators that take `**kwargs` parameter. The keywords are parameter names and values are parameter values. When the function takes multiple parameters, they should specify the order; ordering depends on python 3.5+ behavior that dictionaries are ordered implicitly.\n\nA conditional function is used to modify the values of one or more parameters conditional on the value(s) of other parameters. A conditional function may return a dict with keys being parameter names that should be updated, and values being the new parameter values.\n\nA validation function is intended to check that one or more parameter values conform to certain restrictions. The return value of a validation function is ignored.\n\nBoth conditional and validation functions can throw ValidationError.\n\nThese functions can be associated with parameters in two ways. First, using the `conditionals` and `validations` arguments of the command decorator. These are dicts with a parameter name or tuple of parameter names being the key and the function being the value. Second, validation functions can be associated with parameters when they are decorated with `@autoclick.validation` and the parameter type matches the type argument of the validation decorator. Multi-parameter validations can only be associated via the first method. Since conditionals are expected to be multi-valued, there is no `@autoclick.conditional` annotation, i.e. they must always be explicitly specified.\n\n### Type matching\n\nYou can also use distinct types created by the `typing.NewType` function for type matching validations. For example, if you want to define a parameter that must be positive and even:\n\n```python\nimport autoclick\nfrom typing import NewType\n\nPositiveEven = NewType(\'PositiveEven\', int)\n\n@autoclick.validation(PositiveEven)\ndef validate_positive_even(arg: int):\n  if arg < 0:\n    raise autoclick.ValidationError()\n  if arg % 2 != 0:\n    raise autoclick.ValidationError()\n```\n\nNote that the typing library does not currently provide an intersection type. Thus, Positive, Even, and PositiveEven must all be distinct validations. There are two ways to simplify:\n\n1. Add the parameter to the validation dict of the command decorator with a tuple of mutliple functions as the value:\n\n```python\nimport autoclick\n\n@autoclick.command(\n    validations={\n        "a": positive, even\n    }\n)\n```\n\n2. Create a composite validation:\n\n```python\nimport autoclick\n\n@autoclick.validation(PositiveEven, (positive, even))\ndef validate_positive_even(arg: int):\n  pass\n```\n\nor even\n\n```python\nimport autoclick\nautoclick.validation(PositiveEven, (positive, even))\n```\n\n### Docstring utilization\n\nAutoClick uses the [docparse](https://github.com/jdidion/docparse) library to parse the docstrings of command functions and composites to extract help text. Note that currently docparse only supports Google-style docstrings.\n\n```python\n# test.py\nimport autoclick\n\n@autoclick.command(show_defaults=True)\ndef main(x: str = "hello"):\n    """Print a string\n\n    Args:\n        x: The string to print.\n    """\n    print(x)\n\nif __name__ == "__main__":\n    main()\n```\n\n```bash\n$ python test.py --help\nUsage: test.py [OPTIONS] [X]\n\n  Print a string\n\nOptions:\n  -x, --x TEXT  The string to print.  [default: hello]\n  --help        Show this message and exit.\n```\n\n## Groups\n\nAutoClick CLIs can have multiple subcommands, by creating command groups:\n\n```python\nfrom autoclick import group\n\n@group("myprog")\ndef myprog():\n    pass\n\n@myprog.command("hi")\ndef say_hi(name: str):\n    print(f"hello {name}")\n    \n@myprog.command("bye")\ndef say_bye(name: str):\n    print(f"byebye {name}")\n```\n\nAutoClick provides alternative group types. For example, `DefaultAutoClickGroup` can take a default command name to run when a command is not specified:\n\n```python\nfrom autoclick import group\nfrom autoclick.commands import DefaultAutoClickGroup\n\n@group(\n    "myprog",\n    group_class=DefaultAutoClickGroup,\n    default="hello"\n)\ndef myprog():\n    pass\n\n# This command is run by default if the command name is not specified\ndef hello(name: str):\n    print(f"hi {name}")\n```\n\n## Installation\n\n```bash\n$ pip intall autoclick\n```\n\n## Runtime Dependencies\n\n* Python 3.6+\n* docparse\n\n## Build dependencies\n\n* poetry 0.12+\n* pytest (with pytest-cov plugin)\n\n## Details\n\n### Option attribute inference\n\nThe following sections describe details of how the arguments to click classes/functions are inferred from the type and docstring information:\n\n#### All Parameters\n\n* name (long): parameter name; underscores converted to dashes unless keep_underscores=True in the command decorator.\n* name (short): starting from the left-most character of the parameter name, the first character that is not used by another parameter or by any built-in; can be overridden by specifying the \'parameter_names\' dictionary in the command decorator.\n* type: inferred from the type hint; if type hint is missing, inferred from the default value; if default value is missing, str.\n* required: by default, true for positional arguments (Arguments) and false for keyword arguments (Options); if positionals_as_options=True in the command decorator, positional arguments are instead required Options. Required keyword arguments can be specified in the \'required\' list in the command decorator.\n* default: unset for positional arguments, keyword value for keyword arguments.\n* nargs: 1 unless type is Tuple (in which case nargs is the number of arguments to the Tuple).\n\n#### Option-only\n\n* hide_input: False unless the command \'hidden\' parameter is specified and includes the parameter name.\n* is_flag: True for keyword arguments of type boolean; assumed to be the True option unless the name starts with \'no\'; the other option will always be inferred by adding/removing \'no-\'\n* multiple: True for sequence types\n* help: Parsed from docstring.\n\n## Todo\n\n* Documentation for positional arguments (see [](https://github.com/pallets/click/issues/587))\n* Handle return values (e.g. if a int, treat as a return code; if a dict, convert to JSON and write to stdout, etc)\n* Look at incorporating features from contributed packages: [](https://github.com/click-contrib)',
    'author': 'John Didion',
    'author_email': 'github@didion.net',
    'url': 'https://github.com/jdidion/autoclick',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from abc import ABCMeta, abstractmethod
from enum import Enum, EnumMeta
from numbers import Number
import pathlib
import re
from typing import (
    Callable,
    Dict,
    Generic,
    Iterable,
    Match,
    NewType,
    Optional,
    Pattern,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast
)

import click

from autoclick.types import DEFAULT_METAVAR, AggregateTypeMixin, autoconversion


T = TypeVar("T")

ReadablePath = NewType("ReadablePath", pathlib.Path)
ReadableFile = NewType("ReadableFile", pathlib.Path)
ReadableDir = NewType("ReadableDir", pathlib.Path)

WritablePath = NewType("WritablePath", pathlib.Path)
WritableFile = NewType("WritableFile", pathlib.Path)
WritableDir = NewType("WritableDir", pathlib.Path)


class BaseType(click.ParamType):
    def __init__(self, metavar: str = None):
        self.metavar = metavar

    def get_metavar(self, param):
        return self.metavar or DEFAULT_METAVAR


class Directory(BaseType):
    name = "dir"

    def __init__(
        self, *args, create: bool = False, metavar: str = "DIR", **kwargs
    ):
        super().__init__(metavar)
        kwargs.update(
            file_okay=False, dir_okay=True, readable=True, writable=True
        )
        self._path_type = click.types.Path(*args, **kwargs)
        self._create = create

    def convert(self, value, param, ctx) -> pathlib.Path:
        path = pathlib.Path(self._path_type(value, param, ctx))
        if self._create and not path.exists():
            path.mkdir(parents=True)
        return path


class OptionalTuple(click.types.Tuple):
    def __init__(self, types):
        super().__init__([])
        self.types = types

    def __call__(self, value, param=None, ctx=None):
        if value is None or all(v is None for v in value):
            return None
        return super().__call__(value, param, ctx)


class DelimitedList(BaseType):
    name = "list"

    def __init__(
        self, item_type: Callable[[str], T] = str, delimiter=",",
        strip_whitespace=True, choices: Optional[Iterable[T]] = None,
        metavar: str = "LIST"
    ):
        super().__init__(metavar)
        self.item_type = item_type
        self.delimiter = delimiter
        self.strip_whitespace = strip_whitespace
        self.choice = click.Choice(list(choices)) if choices else None

    def convert(self, value, param, ctx) -> Tuple[T]:
        if not value:
            return cast(Tuple[T], ())
        items = value.split(self.delimiter)
        if self.strip_whitespace:
            items = (item.strip() for item in items)
        items = tuple(self.item_type(item) for item in items)
        if self.choice:
            items = tuple(self.choice.convert(i, param, ctx) for i in items)
        return items


class RegExp(BaseType, metaclass=ABCMeta):
    name = "regexp"

    def __init__(
        self, pattern: Union[str, Pattern], exact: bool = True, metavar: str = "PATTERN"
    ):
        super().__init__(metavar)
        if isinstance(pattern, str):
            self.pattern = re.compile(pattern)
        else:
            self.pattern = cast(Pattern, pattern)
        self.exact = exact

    def convert(self, value, param, ctx):
        if self.exact:
            match = self.pattern.match(value)
        else:
            match = self.pattern.search(value)
        return self._handle_match(match, value, param, ctx)

    @abstractmethod
    def _handle_match(self, match: Match, value, param, ctx):
        pass


class Matches(RegExp):
    """Returns True if the regular expression matches the value.
    """
    def _handle_match(self, match: Match, value, param, ctx):
        return match is not None


class Parse(RegExp):
    """Uses regular expression to parse a value and returns the capture groups
    as a tuple.

    Raises:
        BadParameter if the regular expression does not match the value.
    """
    def _handle_match(self, match, value, param, ctx):
        if match is None:
            self.fail("Pattern {} does not match value {}".format(
                self.pattern, value), ctx=ctx, param=param)
        return match.groups()


E = TypeVar("E", bound=EnumMeta)


@autoconversion(Enum)
class EnumChoice(Generic[E], BaseType):
    """Translates string values into enum instances.

    Args:
        enum_class: Callable (typically a subclass of Enum) that returns enum
            instances.
        xform: How to transform string values before passing to callable;
            upper = convert to upper case; lower = convert to lower case;
            None = don't convert. (default = upper)
    """
    name = "choice"

    def __init__(
        self, enum_class: Sequence[E], xform="upper", metavar: Optional[str] = None
    ):
        super().__init__(metavar)
        self.choice = click.Choice(list(e.name for e in list(enum_class)))
        self.metavar = metavar or self.choice.get_metavar(None)
        self.enum_class = enum_class
        if xform == "upper":
            self.xform = str.upper
        elif xform == "lower":
            self.xform = str.lower
        else:
            self.xform = lambda s: s

    def convert(self, value, param, ctx) -> E:
        if isinstance(value, str):
            return self.enum_class[self.choice.convert(self.xform(value), param, ctx)]
        else:
            return cast(E, value)

    def get_missing_message(self, param):
        return self.choice.get_missing_message(param)

    def __repr__(self):
        return str(self.choice)


K = TypeVar("K")
V = TypeVar("V")


@autoconversion(dict, pass_type=False)
class Mapping(BaseType, AggregateTypeMixin):
    """
    """
    name = "mapping"

    def __init__(
        self, kv_types: Tuple[Type[K], Type[V]], metavar: Optional[str] = None
    ):
        super().__init__(metavar)
        self.kv_types = kv_types

    def convert(self, value, param, ctx) -> Tuple[K, V]:
        # Todo: support conversion of key and value types. Need to expose
        #   core.CONVERSIONS.
        k, v = value.split("=")
        return self.kv_types[0](k), self.kv_types[1](v)

    def aggregate(self, values: Iterable[Tuple[K, V]]) -> Dict[K, V]:
        return dict(values)


N = TypeVar('N', bound=Number)


class BaseNumericType(Generic[N], BaseType):
    def __init__(
        self,
        datatype: Callable[..., Optional[N]],
        name: str = None,
        metavar: str = None
    ):
        self.datatype = datatype
        type_name = datatype.__name__.upper()
        self.name = name or self._get_default_name(type_name)
        super().__init__(metavar=metavar or type_name)

    def _get_default_name(self, type_name: str) -> str:
        pass

    def convert(self, value, param, ctx) -> Optional[N]:
        try:
            return self.datatype(value)
        except (ValueError, UnicodeError):
            self.fail(f'{value} is not a valid {self.datatype}', param, ctx)


class Positive(Generic[N], BaseNumericType[N]):
    def _get_default_name(self, type_name: str) -> str:
        return f'POSITIVE_{type_name}'

    def convert(self, value, param, ctx) -> Optional[N]:
        value = super().convert(value, param, ctx)
        if value is not None and value < self.datatype(0):
            self.fail(f'{self.datatype} value must be >= 0', param, ctx)
        return value


class Range(Generic[N], BaseNumericType[N]):
    def __init__(
        self,
        datatype: Callable[..., Optional[N]],
        min_value=None,
        max_value=None,
        **kwargs
    ):
        self.min_value = min_value
        self.max_value = max_value
        if datatype is None:
            datatype = type(min_value)
        super().__init__(datatype, **kwargs)

    def _get_default_name(self, type_name: str) -> str:
        return f'RANGE_{self.datatype}'

    def convert(self, value, param, ctx) -> Optional[N]:
        value = super().convert(value, param, ctx)

        failures = []

        if self.min_value is not None and value < self.min_value:
            failures.append(f'must be >= {self.min_value}')

        if self.max_value is not None and value > self.max_value:
            failures.append(f'must be <= {self.max_value}')

        if failures:
            self.fail(f'{value}' + ' and '.join(failures))

        return value
